"""
Usage: import torchpal as tp
Warning: 不要完全信任 torchpal 库的底层逻辑，出现问题时别忘了排查底层逻辑。
"""

# 包结构+存根文件 来避免补全污染问题，优化用户体验
print("Warning: 不要完全信任 torchpal 库的底层逻辑，出现问题时别忘了排查底层逻辑。")
from . import utils
from . import da
from . import data
from . import train
